# Autogenerated by Astropy's setup.py on 2017-12-13 17:48:11
from __future__ import unicode_literals
import datetime

version = "2.0.3"
githash = "4ea556e14340760bd58e53ea2cbccde7ce5567c3"


major = 2
minor = 0
bugfix = 3

release = True
timestamp = datetime.datetime(2017, 12, 13, 17, 48, 11)
debug = False

try:
    from ._compiler import compiler
except ImportError:
    compiler = "unknown"

try:
    from .cython_version import cython_version
except ImportError:
    cython_version = "unknown"
